/*
 * Copyright 2021 Free Software Foundation, Inc.
 *
 * This file is part of GNU Radio
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 */

/***********************************************************************************/
/* This file is automatically generated using bindtool and can be manually edited  */
/* The following lines can be configured to regenerate this file during cmake      */
/* If manual edits are made, the following tags should be modified accordingly.    */
/* BINDTOOL_GEN_AUTOMATIC(0)                                                       */
/* BINDTOOL_USE_PYGCCXML(0)                                                        */
/* BINDTOOL_HEADER_FILE(decoder.h)                                        */
/* BINDTOOL_HEADER_FILE_HASH(079fc56f59ff41939a36542553e3e79e)                     */
/***********************************************************************************/

#include <pybind11/complex.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

namespace py = pybind11;

#include <rds/decoder.h>
// pydoc.h is automatically generated in the build directory
#include <decoder_pydoc.h>

void bind_decoder(py::module& m)
{

    using decoder    = ::gr::rds::decoder;


    py::class_<decoder, gr::sync_block, gr::block, gr::basic_block,
        std::shared_ptr<decoder>>(m, "decoder", D(decoder))

        .def(py::init(&decoder::make),
           py::arg("log"),
           py::arg("debug"),
           D(decoder,make)
        )
        



        ;




}








