/*
 * Copyright 2021 Free Software Foundation, Inc.
 *
 * This file is part of GNU Radio
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 */

/***********************************************************************************/
/* This file is automatically generated using bindtool and can be manually edited  */
/* The following lines can be configured to regenerate this file during cmake      */
/* If manual edits are made, the following tags should be modified accordingly.    */
/* BINDTOOL_GEN_AUTOMATIC(0)                                                       */
/* BINDTOOL_USE_PYGCCXML(0)                                                        */
/* BINDTOOL_HEADER_FILE(parser.h)                                        */
/* BINDTOOL_HEADER_FILE_HASH(b439a5ab6aa91a4cc6ad66b2105f9900)                     */
/***********************************************************************************/

#include <pybind11/complex.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

namespace py = pybind11;

#include <rds/parser.h>
// pydoc.h is automatically generated in the build directory
#include <parser_pydoc.h>

void bind_parser(py::module& m)
{

    using parser    = ::gr::rds::parser;


    py::class_<parser, gr::block, gr::basic_block,
        std::shared_ptr<parser>>(m, "parser", D(parser))

        .def(py::init(&parser::make),
           py::arg("log"),
           py::arg("debug"),
           py::arg("pty_locale"),
           D(parser,make)
        )
        




        
        .def("reset",&parser::reset,       
            D(parser,reset)
        )

        ;




}








