/*
 * Copyright 2021 Free Software Foundation, Inc.
 *
 * This file is part of GNU Radio
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 */

/***********************************************************************************/
/* This file is automatically generated using bindtool and can be manually edited  */
/* The following lines can be configured to regenerate this file during cmake      */
/* If manual edits are made, the following tags should be modified accordingly.    */
/* BINDTOOL_GEN_AUTOMATIC(0)                                                       */
/* BINDTOOL_USE_PYGCCXML(0)                                                        */
/* BINDTOOL_HEADER_FILE(encoder.h)                                        */
/* BINDTOOL_HEADER_FILE_HASH(baad12f97f57978fc0a3f0b5fec44341)                     */
/***********************************************************************************/

#include <pybind11/complex.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

namespace py = pybind11;

#include <rds/encoder.h>
// pydoc.h is automatically generated in the build directory
#include <encoder_pydoc.h>

void bind_encoder(py::module& m)
{

    using encoder    = ::gr::rds::encoder;


    py::class_<encoder, gr::sync_block, gr::block, gr::basic_block,
        std::shared_ptr<encoder>>(m, "encoder", D(encoder))

        .def(py::init(&encoder::make),
           py::arg("pty_locale"),
           py::arg("pty"),
           py::arg("ms"),
           py::arg("ps"),
           py::arg("af1"),
           py::arg("tp"),
           py::arg("ta"),
           py::arg("pi_country_code"),
           py::arg("pi_coverage_area"),
           py::arg("pi_reference_number"),
           py::arg("radiotext"),
           D(encoder,make)
        )
        




        
        .def("set_ps",&encoder::set_ps,       
            py::arg("ps"),
            D(encoder,set_ps)
        )

        ;




}








